THIS PATCH HAS BEEN MODIFIED/BACKPORTED FROM THE ORIGINAL 2.6.20 PATCH

Re: [uml-devel] [UML] fix crash in block layer

Jason Lunz
Fri, 16 Feb 2007 09:02:30 -0800

On Thu, Feb 15, 2007 at 08:29:28PM +0100, Blaisorblade wrote:
> The following could be a suggestion, if max_low_pfn is not used between the 
> old and the new moment of assignment (and it seems it is not). This is just 
> an idea however:
> 
> mem_init:
> 
> -        max_low_pfn = ...
>         /* this will put all low memory onto the freelists */
>         totalram_pages = free_all_bootmem();
> +        max_low_pfn = totalram_pages;
> #ifdef CONFIG_HIGHMEM
>         totalhigh_pages = highmem >> PAGE_SHIFT;
>         totalram_pages += totalhigh_pages;
> #endif
>         num_physpages = totalram_pages;
>         max_pfn = totalram_pages;
>
> Please note that I did not spend a lot of time on this, so everything
> could be wrong. However, testing cannot help with uml_reserved
> handling, and this is a dark corner.  So things should be better
> understood before merging the patch.

I agree - I have only a vague idea about what uml_reserved means.

> The code is too convoluted for a brief look - drawing a picture which
> explains all those variables would help. Both for UML and for every
> arch...

blk_queue_bounce_limit() calls init_emergency_isa_pool() to get dma-zone
pages to use as bounce buffers when its caller passes a dma_addr limit
that is less than max_low_pfn. The BLK_BOUNCE_ANY macro is supposed to
mean "never bounce", and it's defined as:

#define BLK_BOUNCE_ANY          ((u64)blk_max_pfn << PAGE_SHIFT)

So it presumes that max_pfn >= max_low_pfn. uml's mem_init() is
violating this assumption - when uml_reserved is subtracted from
max_pfn, we end up with max_pfn < max_low_pfn, so BLK_BOUNCE_ANY has the
opposite of the intended effect. blk_queue_bounce_limit therefore tries
to create a mempool with zone-dma pages on a no-dma-zone arch and the
kernel goes BUG().

So I think your idea is correct. It passes my testing - I can still use
lvm within uml. I have not tested CONFIG_HIGHMEM, but here's an
implementation against 2.6.20.

Jeff, please drop my other patch and use this one.

Signed-off-by: Jason Lunz <[EMAIL PROTECTED]>

---
 arch/um/kernel/mem.c |    3 +--
 1 file changed, 1 insertion(+), 2 deletions(-)

Index: linux-2.6.20-uml/arch/um/kernel/mem.c
===================================================================
--- linux-2.6.20-uml.orig/arch/um/kernel/mem.c
+++ linux-2.6.20-uml/arch/um/kernel/mem.c
@@ -63,8 +63,6 @@
 
 void mem_init(void)
 {
-	max_low_pfn = (high_physmem - uml_physmem) >> PAGE_SHIFT;
-
         /* clear the zero-page */
         memset((void *) empty_zero_page, 0, PAGE_SIZE);
 
@@ -79,6 +77,7 @@
 
 	/* this will put all low memory onto the freelists */
 	totalram_pages = free_all_bootmem();
+	max_low_pfn = totalram_pages;
 	totalhigh_pages = highmem >> PAGE_SHIFT;
 	totalram_pages += totalhigh_pages;
 	num_physpages = totalram_pages;

